/*******************************************************************************
  @doc public app_rescue

  Copyright (C) 2008 - 2011 Garant Plus Technologies, All Rights Reserved

  Asterisk to Rescue-4E1 control interface.

  @module
    app_rescue | asterisk to rescue control interface implementation.

  @contents2
    app_rescue.c |


 * This program is free software, distributed under the terms of
 * the GNU General Public License
*******************************************************************************/

#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>

#include "asterisk.h"
#include "asterisk/module.h"
#include "asterisk/logger.h"
#include "asterisk/config.h"

#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/ip.h> 
#include <fcntl.h>
#include <arpa/inet.h>
#include <signal.h>

#include <dahdi/user.h>


#ifndef AST_MODULE
	#define AST_MODULE "app_rescue"
#endif

#define TEXT_NAME     "app_rescue"
#define TEXT_VERSION  "1.0.2"
#define TEXT_BUILD    "Build: " __DATE__ " " __TIME__
#define TEXT_HELLO    "Application " TEXT_NAME " version " TEXT_VERSION ", " TEXT_BUILD "\n"
#define TEXT_GOODBYE  "Application " TEXT_NAME " unloaded.\n"

static const char *app_name = "rescue";
static const char *synopsis = "";
static const char *description = TEXT_HELLO"interface with rescue board.\n";
static const char *config_file_name = "/etc/asterisk/rescue.conf";

// config and defaults
static char cmd_local_addr[128] = { "0.0.0.0" };
static int cmd_port = 10000;
static unsigned read_timeout_secs = 1;
static int async_mode = 0;

//
static int cmd_sock = -1;
static int ctl = -1;

// 1 - Up, 0 - Down, -1 - Not exists
static int spans_status[DAHDI_MAX_SPANS];
// protect spans_status[]
static ast_mutex_t mtx; 


/*******************************************************************************
@func
Reads dahdi span status.
@rval
Returns: 
      NULL on error or string representing spans state.
      result will contain:
        0 - span DOWN
        1 - span UP
       -1 - span is not configured
*******************************************************************************/
static char* getalarms(int span, int *result)
{
	int res;
	static char tmp[256];
	char alarms[50];
	static struct dahdi_spaninfo spans[DAHDI_MAX_SPANS];

	*result = 0;
	spans[span].spanno = span;

	res = ioctl(ctl, DAHDI_SPANSTAT, &spans[span]);
	if (res) {
		*result = -1;
		//if (err) fprintf(stderr, "Unable to get span info on span %d: %s\n", span, strerror(errno));
		return NULL;
	}

	strcpy(alarms, "");
	if (spans[span].alarms > 0) {
		if (spans[span].alarms & DAHDI_ALARM_BLUE)
			strcat(alarms,"BLU/");
		if (spans[span].alarms & DAHDI_ALARM_YELLOW)
			strcat(alarms, "YEL/");
		if (spans[span].alarms & DAHDI_ALARM_RED)
			strcat(alarms, "RED/");
		if (spans[span].alarms & DAHDI_ALARM_LOOPBACK)
			strcat(alarms,"LB/");
		if (spans[span].alarms & DAHDI_ALARM_RECOVER)
			strcat(alarms,"REC/");
		if (spans[span].alarms & DAHDI_ALARM_NOTOPEN)
			strcat(alarms, "NOP/");
		if (!strlen(alarms))
			strcat(alarms, "UUU/");
		if (strlen(alarms)) {
			/* Strip trailing / */
			alarms[strlen(alarms)-1]='\0';
		}
	} else {
		if (spans[span].numchans) {
			strcpy(alarms, "OK");
			*result = 1;
		}
		else {
			strcpy(alarms, "UNCONFIGURED");
		}
	}

	snprintf(tmp, sizeof(tmp), "%15s %s", alarms, spans[span].desc);
	return tmp;
}


/*******************************************************************************
@func
Reads dahdi spans status.
@rval
Returns: 
*******************************************************************************/
static void read_spans_status(int async, int print)
{
	int x;
	int r[DAHDI_MAX_SPANS];

	memset(r, 0, sizeof(r));

	for (x = 0; x < DAHDI_MAX_SPANS; x++) {
		char *s = getalarms(x, r + x);
		if (s && print) {
			ast_log(AST_LOG_DEBUG, "span %i - %i, '%s'.\n", x, spans_status[x], s);
		}
	}

	if(async) ast_mutex_lock(&mtx);

	memcpy(spans_status, r, sizeof(spans_status));

	if (async) ast_mutex_unlock(&mtx);
}

static int stop_span_monitor = 0;
static pthread_t span_monitor_thread = AST_PTHREADT_NULL;


/*******************************************************************************
@func
Reads dahdi spans status in separate thread.
@rval
Returns:
*******************************************************************************/
static void* span_monitor_thread_fn(void *data)
{
	while (!stop_span_monitor) {
		int i = 0;

		for (i = 0; i < read_timeout_secs; i++) {
			if (stop_span_monitor)
				break;
			sleep(1);
		}

		if (stop_span_monitor)
			break;

		read_spans_status(1, 0);
	}
	return 0;
}


/*******************************************************************************
@func
Starts dahdi spans state monitor.
@rval
Returns:
0 - success, -1 on error
*******************************************************************************/
static int start_spans_monitor()
{
	stop_span_monitor = 0;
	if (ast_pthread_create_background(&span_monitor_thread, NULL, span_monitor_thread_fn, NULL) < 0) {
		ast_log(AST_LOG_ERROR, "Unable to start span monitor thread.\n");
		return -1;
	}
	return 0;
}


/*******************************************************************************
@func
Stops dahdi spans state monitor.
@rval
Returns:
0 - success, -1 on error
*******************************************************************************/
static void stop_spans_monitor()
{
	stop_span_monitor = 1;

	if (span_monitor_thread && (span_monitor_thread != AST_PTHREADT_STOP) && (span_monitor_thread != AST_PTHREADT_NULL)) {
		pthread_cancel(span_monitor_thread);
		pthread_kill(span_monitor_thread, SIGURG);
		pthread_join(span_monitor_thread, NULL);
	}
	span_monitor_thread = AST_PTHREADT_STOP;
}


static int stop_cmd_iface = 0;
static pthread_t cmd_iface_thread = AST_PTHREADT_NULL;


/*******************************************************************************
@func
Initialize UDP socket
@rval
Returns:
-1 on error, else socket descriptor
*******************************************************************************/
int setup_socket(int localport, int sockettype, int ipproto, const char* local_addr)
{
	struct sockaddr_in sock;

	in_addr_t addr = inet_addr(local_addr);
	int parm/*, res, flags*/;
	int s;

	memset(&sock, 0, sizeof(struct sockaddr_in));

	sock.sin_family = AF_INET;
	sock.sin_port = htons(localport);
	memcpy(&sock.sin_addr, &addr, sizeof(addr));

	s = socket(PF_INET, sockettype, ipproto);
	if (s < 0) {
		ast_log(AST_LOG_ERROR, "Cannot create UDP socket, errno=%d: %s\n", errno, strerror(errno));
		return -1;
	}

	parm = 1;
	setsockopt(s, SOL_SOCKET, SO_REUSEADDR, &parm, sizeof(int));

	/*
	will use blocking recv

	res = fcntl(s, F_GETFL);
	if (res < 0) {
		ast_log(AST_LOG_WARNING, "Could not obtain flags for socket fd: %s.\n", strerror(errno));
		return -1;
	}

	flags = res | O_NONBLOCK;
	res = fcntl(s, F_SETFL, flags);
	if (res < 0) {
		ast_log(AST_LOG_WARNING, "Could not set socket fd non-blocking: %s.\n", strerror(errno));
		return -1;
	}
	*/
	if (bind(s, (struct sockaddr*)&sock, sizeof(sock)) < 0) {
		ast_log(AST_LOG_ERROR, "Cannot bind receiver socket, errno=%d: %s\n", errno, strerror(errno));
		close(s);
		return -1;
	}
	/*
	if (sockettype != SOCK_DGRAM)
		if (listen(s, 8) < 0) {
			ast_log(LOG_ERROR, "Cannot listen on socket, errno=%d: %s\n", errno, strerror(errno));
			close(s);
			return -1;
		}
		*/
	return s;
}


/*******************************************************************************
@func
Process command from Rescue
@rval
Returns:
*******************************************************************************/
static void* cmd_iface_thread_fn(void *data)
{
	int r[sizeof(spans_status)];
	char buffer[1024];
	char answer[(DAHDI_MAX_SPANS + 1) * 30]; // 30 chars per line, more than enough

	const char *listatus = "console listatus";
	const int listatus_len = strlen(listatus);

	while (!stop_cmd_iface) {

		struct sockaddr_in remote;
		int res = 0, i = 0;
		socklen_t remote_len = sizeof(remote);

		buffer[0] = 0;

		// blocking read command from rescue
		res = recvfrom(cmd_sock, buffer, sizeof(buffer), 0, (struct sockaddr*)&remote, &remote_len);

		if (stop_cmd_iface)
			break;

		if (res < 0) {
			sleep(1);
			continue;
		}

		if (res < listatus_len || strcasecmp(listatus, buffer)) {
			sleep(1);
			continue;
		}

		// read status from dahdi
		if (async_mode) {
			ast_mutex_lock(&mtx);
			memcpy(r, spans_status, sizeof(spans_status));
			ast_mutex_unlock(&mtx);
		}
		else {
			read_spans_status(async_mode, 0);
			memcpy(r, spans_status, sizeof(spans_status));
		}

		// generate asnwer
		strcpy(answer, "");
		for (i = 1; i < DAHDI_MAX_SPANS; i++) {
			char tmp[80];

			if (r[i] == -1)
				continue;

			snprintf(tmp, sizeof(tmp) - 1, "Li%i is %s\r\n", i - 1, r[i]?"OK":"DOWN");
			strcat(answer, tmp);
		}
		strcat(answer, "OK listatus\r\nConsole : Complete\r\n");

		ast_log(AST_LOG_DEBUG, "\n%s", answer);

		// send asnwer back
		res = sendto(cmd_sock, answer, strlen(answer), 0, (struct sockaddr*)&remote, remote_len);
		if (res < 0) {
			ast_log(AST_LOG_ERROR, "failed to send answer %i, '%s'", errno, strerror(errno));
		}

		// flood protection, don't ask dahdi too often
		if (!async_mode) 
			sleep(1);
	}
	ast_log(AST_LOG_DEBUG, "Exit cmd thread.\n");
	return 0;
}


/*******************************************************************************
@func
Starts command interface
@rval
Returns:
*******************************************************************************/
static int start_cmd_interface()
{
	cmd_sock = -1;

	cmd_sock = setup_socket(cmd_port, SOCK_DGRAM, IPPROTO_UDP, cmd_local_addr);
	if (cmd_sock < 0)
		return -1;

	stop_cmd_iface = 0;

	if (ast_pthread_create_background(&cmd_iface_thread, NULL, cmd_iface_thread_fn, NULL) < 0) {
		ast_log(AST_LOG_ERROR, "Unable to start cmd interface thread.\n");
		return -1;
	}
	return 0;
}


/*******************************************************************************
@func
Stops command interface
@rval
Returns:
*******************************************************************************/
static void stop_cmd_interface()
{
	stop_cmd_iface = 1;

	if (cmd_sock >= 0)
		close(cmd_sock);

	if (cmd_iface_thread && (cmd_iface_thread != AST_PTHREADT_STOP) && (cmd_iface_thread != AST_PTHREADT_NULL)) {
		pthread_cancel(cmd_iface_thread);
		pthread_kill(cmd_iface_thread, SIGURG);
		pthread_join(cmd_iface_thread, NULL);
	}
	cmd_iface_thread = AST_PTHREADT_STOP;
}


/*******************************************************************************
@func
Load configuration from the file.
@rval
Returns: 0 on error or 1 on success, or -1 if nothing loaded.
*******************************************************************************/
static int load_config()
{
	struct ast_flags config_flags = { 0 };
	struct ast_config *cfg = ast_config_load(config_file_name, config_flags);
	struct ast_variable *var;

	if (!cfg) {
		ast_log(AST_LOG_WARNING, "Unable to load config '%s', using defaults\n", config_file_name);
		return 0;
	}

	if (cfg == CONFIG_STATUS_FILEMISSING) {
		return 0;
	}

	if (cfg == CONFIG_STATUS_FILEINVALID) {
		return -1;
	}

	if (cfg == CONFIG_STATUS_FILEUNCHANGED) {
		return 1;
	}

	for (var = ast_variable_browse(cfg, "general"); var; var = var->next) {

		if (!strcasecmp(var->name, "port")) {
			cmd_port = atoi(var->value);
		}

		else if (!strcasecmp(var->name, "address")) {
			strncpy(cmd_local_addr, var->value, sizeof(cmd_local_addr) - 1);
		}

		else if (!strcasecmp(var->name, "timeout")) {
			read_timeout_secs = atoi(var->value);
			if (read_timeout_secs < 0)
				read_timeout_secs = 1;
		}

		else if (!strcasecmp(var->name, "mode")) {

			if (!strcasecmp(var->value, "sync"))
				async_mode = 0;
			else if (!strcasecmp(var->value, "async"))
				async_mode = 1;
			else
				ast_log(AST_LOG_WARNING, "unrecognized operation mode value '%s', please use 'sync' or 'async'\n", var->value);

			//async_mode = 0; //async not tested yet
		}

		else {
			ast_log(AST_LOG_WARNING, "Unknown parameter: %s\n", var->name);
		}
	}
	return 1;
}


/*******************************************************************************
@func
Asterisk app main, we have nothing to do here.
@rval
Returns: 
*******************************************************************************/
static int app_main(struct ast_channel*chan, const char*data)
{
	return 0;
}


/*******************************************************************************
@func
Called by asterisk to load our module
@rval
Returns: AST_MODULE_LOAD_DECLINE on errors, AST_MODULE_LOAD_SUCCESS on success
*******************************************************************************/
static int load_module(void)
{
	ast_log(AST_LOG_NOTICE, TEXT_HELLO);

	ast_mutex_init(&mtx);
	memset(spans_status, 0xff, sizeof(spans_status));

	ctl = open("/dev/dahdi/ctl", O_RDWR);
	if (ctl < 0) {
		ast_log(LOG_ERROR, "Unable to open /dev/dahdi/ctl: %s\n", strerror(errno));
		return AST_MODULE_LOAD_DECLINE;
	}

	load_config();

	if (async_mode) {
		if (start_spans_monitor()) {
			close(ctl);
			ctl = -1;
			return AST_MODULE_LOAD_DECLINE;
		}
	}

	if (start_cmd_interface()) {
		stop_spans_monitor();
		close(ctl);
		ctl = -1;
		return AST_MODULE_LOAD_DECLINE;
	}

	return ast_register_application(app_name, app_main, synopsis, description) ? AST_MODULE_LOAD_DECLINE : AST_MODULE_LOAD_SUCCESS;
}


/*******************************************************************************
@func
Called by asterisk to unload our app
@rval
Returns: 
*******************************************************************************/
static int unload_module(void)
{
	stop_cmd_interface();

	if (async_mode)
		stop_spans_monitor();

	if(ctl >= 0)
		close(ctl);

	ast_log(AST_LOG_NOTICE, TEXT_GOODBYE);

	return ast_unregister_application(app_name);
}


/*******************************************************************************
@func
Called by asterisk to reload our app
@rval
Returns: AST_MODULE_LOAD_DECLINE on errors, AST_MODULE_LOAD_SUCCESS on success
*******************************************************************************/
static int reload(void)
{
	unload_module();
	return load_module();
}


/*******************************************************************************
@
Asterisk app declaration
*******************************************************************************/
AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_DEFAULT, "Asterisk to Rescue interface",
	.load = load_module,
	.unload = unload_module,
	.reload = reload,
	);



/*******************************************************************************

  Copyright (C) GarantPlus Technologies Ltd., 2009-2017


  All rights reserved. No part of this computer program may be reproduced,
  stored in any retrieval system, or transmitted, in any form or by any
  means, electronic, mechanical, photocopying, recording, or otherwise,
  without the prior written permission of Odin TeleSystems Inc., or without
  a permission stipulated in the agreement/contract under which this computer
  program was supplied.

*******************************************************************************/
